from http import HTTPStatus
from typing import Any, Union

from tenacity import (retry, retry_if_exception_type, retry_if_result,
                      stop_after_attempt, wait_exponential)

from featureflags.openapi.config import AuthenticatedClient, Client
from featureflags.openapi.config.types import UNSET, Response, Unset

from .openapi.config.api.client.authenticate import \
    sync_detailed as authenticate
from .openapi.config.api.client.get_all_segments import \
    sync_detailed as retrieve_segments
from .openapi.config.api.client.get_feature_config import \
    sync_detailed as retrieve_flags
from .openapi.config.api.client.get_feature_config_by_identifier import \
    sync_detailed as retrieve_flag_by_identifier
from .openapi.config.api.client.get_segment_by_identifier import \
    sync_detailed as retrieve_segment_by_identifier
from .openapi.config.errors import UnexpectedStatus
from .openapi.config.models import (AuthenticationRequest,
                                    AuthenticationResponse, FeatureConfig,
                                    Segment)
from .sdk_logging_codes import (warn_auth_retying,
                                warning_fetch_all_features_retrying,
                                warning_fetch_all_segments_retrying,
                                warning_fetch_feature_by_id_retrying,
                                warning_fetch_group_by_id_retrying)

TARGET_SEGMENT_RULES_PARAM = "v2"


class UnrecoverableRequestException(Exception):
    def __init__(self, status_code, content):
        self.status_code = status_code
        self.content = content

    def __str__(self):
        return f'Request failed with unrecoverable error: ' \
               f'status_code={self.status_code}, contents={self.content}'


def default_retry_strategy(before_sleep_func=None, on_retry_error=None):
    max_retry_attempts = 10
    return retry(
        retry=(
            retry_if_result(handle_http_result) |
            retry_if_exception_type(UnexpectedStatus)),

        wait=wait_exponential(multiplier=1, max=10),
        before_sleep=before_sleep_func,
        stop=stop_after_attempt(max_retry_attempts),
        retry_error_callback=lambda response: on_retry_error(
            response),
    )


def handle_http_result(response):
    retryable_codes = {HTTPStatus.BAD_GATEWAY, HTTPStatus.NOT_FOUND,
                       HTTPStatus.INTERNAL_SERVER_ERROR}
    code = response.status_code
    if code == HTTPStatus.OK:
        return False

    if code in retryable_codes:
        return True
    else:
        raise UnrecoverableRequestException(response.status_code,
                                            response.content)


def handle_retries_exceeded(retry_state):
    content = retry_state.outcome.result().content
    if content == b'':
        content = ""
    raise UnrecoverableRequestException(
        retry_state.outcome.result().status_code,
        content)


def make_log_warning_before_sleep(warning_fun):
    def log_warning_before_sleep(retry_state):
        error_message = build_retry_warning_message(retry_state)
        warning_fun(retry_state.attempt_number, error_message)

    return log_warning_before_sleep


def build_retry_warning_message(retry_state):
    # Defensive check in case getting result of outcome throws
    try:
        result = retry_state.outcome.result()
        status_code = result.status_code
        content = result.content if result.content != b'' else ""
        error_message = f"status_code={status_code}, content={content}"
    except Exception as e:
        status_code = None
        content = str(e)
        error_message = f"status_code={status_code}, content={content}"
    return error_message


@default_retry_strategy(
    before_sleep_func=make_log_warning_before_sleep(warn_auth_retying),
    on_retry_error=handle_retries_exceeded)
def retryable_authenticate(
        client: Union[AuthenticatedClient, Client],
        body: AuthenticationRequest) -> \
        Response[Union[AuthenticationResponse, Any]]:
    response = authenticate(client=client, body=body)
    return response


@default_retry_strategy(
    before_sleep_func=make_log_warning_before_sleep(
        warning_fetch_all_segments_retrying),
    on_retry_error=handle_retries_exceeded)
def retryable_retrieve_segments(environment_uuid: str,
                                client: AuthenticatedClient,
                                cluster: Union[Unset, str] = UNSET) -> \
        Response[list[Segment]]:
    return retrieve_segments(client=client,
                             environment_uuid=environment_uuid,
                             cluster=cluster,
                             rules=TARGET_SEGMENT_RULES_PARAM)


@default_retry_strategy(
    before_sleep_func=make_log_warning_before_sleep(
        warning_fetch_all_features_retrying),
    on_retry_error=handle_retries_exceeded)
def retryable_retrieve_feature_config(environment_uuid: str,
                                      client: AuthenticatedClient,
                                      cluster: Union[Unset, str] = UNSET) -> \
        Response[list[FeatureConfig]]:
    return retrieve_flags(client=client,
                          environment_uuid=environment_uuid,
                          cluster=cluster)


@default_retry_strategy(
    before_sleep_func=make_log_warning_before_sleep(
        warning_fetch_feature_by_id_retrying),
    on_retry_error=handle_retries_exceeded)
def retryable_retrieve_feature_config_by_identifier(environment_uuid: str,
                                                    identifier: str,
                                                    client:
                                                    AuthenticatedClient,
                                                    cluster: Union[
                                                        Unset, str] = UNSET) \
        -> Response[FeatureConfig]:
    return retrieve_flag_by_identifier(client=client,
                                       identifier=identifier,
                                       environment_uuid=environment_uuid,
                                       cluster=cluster)


@default_retry_strategy(
    before_sleep_func=make_log_warning_before_sleep(
        warning_fetch_group_by_id_retrying),
    on_retry_error=handle_retries_exceeded)
def retryable_retrieve_segment_by_identifier(environment_uuid: str,
                                             identifier: str,
                                             client:
                                             AuthenticatedClient,
                                             cluster: Union[
                                                 Unset, str] = UNSET) \
        -> Response[Segment]:
    return retrieve_segment_by_identifier(client=client,
                                          identifier=identifier,
                                          environment_uuid=environment_uuid,
                                          cluster=cluster,
                                          rules=TARGET_SEGMENT_RULES_PARAM)



import time
from concurrent.futures import Future
from threading import Event, Thread
from typing import Dict

from featureflags.repository import DataProviderInterface

from .api import (UnrecoverableRequestException,
                  retryable_retrieve_feature_config,
                  retryable_retrieve_segments)
from .config import Config
from .openapi.config import AuthenticatedClient
from .sdk_logging_codes import (info_poll_ran_successfully, info_poll_started,
                                info_polling_stopped, info_sdk_init_ok,
                                warn_failed_init_fetch_error,
                                warning_fetch_all_features_failed,
                                warning_fetch_all_groups_failed)
from .util import log


class RetrievalError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"RetrievalError: {self.message}"


class PollingProcessor(Thread):
    __initialised_failed_reason: Dict[bool, str]

    def __init__(self, client: AuthenticatedClient, config: Config,
                 environment_id: str, wait_for_initialization: Event,
                 initialised_failed_reason: Dict[bool, str],
                 ready: Event, stream_ready: Event,
                 repository: DataProviderInterface,
                 cluster: str) -> None:
        Thread.__init__(self)
        self.daemon = True
        self.__environment_id = environment_id
        self.__client = client
        self.__config = config
        self.__running = False
        self.__wait_for_initialization = wait_for_initialization
        self.__initialised_failed_reason = initialised_failed_reason
        self.__ready = ready
        self.__stream_ready = stream_ready
        self.__repository = repository
        self.__cluster = cluster

    def run(self):
        if not self.__running:
            if self.__config.pull_interval < 60:
                log.warning("Pull Interval must be greater than or equal to "
                            "60 seconds, was: " +
                            str(self.__config.pull_interval) +
                            " setting to 60")
                self.__config.pull_interval = 60

            self.__running = True
            #  Get initial flags and groups
            try:
                log.info("Fetching initial target segments and flags")
                self.retrieve_flags_and_segments()
                log.info("Initial target segments and flags fetched")
                if not self.__config.enable_stream:
                    info_poll_started(self.__config.pull_interval)

                #  Segments and flags have been cached so
                #  mark the Client as initialised.
                self.__wait_for_initialization.set()
                info_sdk_init_ok()

            except RetrievalError as ex:
                warn_failed_init_fetch_error(ex)
                self.__initialised_failed_reason[True] = str(ex)
                # Unblock the thread and log that initialization has failed
                self.__wait_for_initialization.set()

            except Exception as ex:
                warn_failed_init_fetch_error(ex)
                self.__initialised_failed_reason[True] = str(ex)
                # Unblock the thread and log that initialization has failed
                self.__wait_for_initialization.set()

            if not self.__config.enable_stream:
                # Sleep for an interval before going into the polling loop,
                # as we've just fetched flags/groups on init.
                time.sleep(self.__config.pull_interval)

            while self.__running:
                start_time = time.time()
                try:
                    if self.__config.enable_stream and \
                            self.__stream_ready.is_set():
                        #  Block until ready.set() is called
                        self.__ready.wait()
                        # on stream disconnect, make sure flags are in sync
                        self.retrieve_flags_and_segments()
                        info_poll_ran_successfully()
                        # Reset the start time so we don't do another poll
                        # immediately
                        start_time = time.time()
                    else:
                        self.retrieve_flags_and_segments()
                        info_poll_ran_successfully()
                        self.__ready.set()
                except RetrievalError as ex:
                    log.error('Polling error: %s',
                              ex)
                except Exception as e:
                    log.exception(
                        'Error: Exception encountered when polling flags. %s',
                        e
                    )

                elapsed = time.time() - start_time
                if elapsed < self.__config.pull_interval:
                    log.debug("Poller sleeping for " +
                              (
                                  self.__config.pull_interval -
                                  elapsed).__str__())
                    " seconds"
                    time.sleep(self.__config.pull_interval - elapsed)

    def stop(self):
        self.__running = False
        info_polling_stopped("Client was closed")

    def retrieve_flags_and_segments(self):
        flags_future = Future()
        segments_future = Future()

        flags_thread = Thread(target=self.__retrieve_flags,
                              args=(flags_future,))
        segments_thread = Thread(target=self.__retrieve_segments,
                                 args=(segments_future,))

        flags_thread.start()
        segments_thread.start()

        flags_thread.join()
        segments_thread.join()

        flags_exception = flags_future.exception()
        segments_exception = segments_future.exception()

        if flags_exception:
            raise flags_exception

        if segments_exception:
            raise segments_exception

    def __retrieve_flags(self, future: Future):
        try:
            log.debug("Loading feature flags")
            flags = retryable_retrieve_feature_config(
                client=self.__client, environment_uuid=self.__environment_id,
                cluster=self.__cluster
            ).parsed
            log.debug("Feature flags loaded")
            for flag in flags:
                log.debug("Put flag %s into repository", flag.feature)
                self.__repository.set_flag(flag)
            future.set_result(
                "Success")

        except UnrecoverableRequestException as e:
            warning_fetch_all_features_failed(e)
            future.set_exception(
                RetrievalError(
                    f"Failed to retrieve flags '{e}'"))

        except Exception as e:
            future.set_exception(
                RetrievalError(
                    f"Failed to retrieve flags '{e}'"))

    def __retrieve_segments(self, future: Future):
        try:
            log.debug("Loading target segments")
            segments = retryable_retrieve_segments(
                client=self.__client,
                environment_uuid=self.__environment_id,
                cluster=self.__cluster
            ).parsed
            log.debug("Target segments loaded")
            for segment in segments:
                log.debug("Put %s segment into repository", segment.identifier)
                self.__repository.set_segment(segment)
            future.set_result(
                "Success")

        except UnrecoverableRequestException as e:
            warning_fetch_all_groups_failed(e)
            future.set_exception(
                RetrievalError(
                    f"Failed to retrieve segments '{e}'"))

        except Exception as ex:
            future.set_exception(
                RetrievalError(
                    f"Failed to retrieve segments '{ex}'"))



"""Contains all the data models used in inputs/outputs"""

from .authenticate_proxy_key_body import AuthenticateProxyKeyBody
from .authentication_request import AuthenticationRequest
from .authentication_request_target import AuthenticationRequestTarget
from .authentication_request_target_attributes import \
    AuthenticationRequestTargetAttributes
from .authentication_response import AuthenticationResponse
from .clause import Clause
from .distribution import Distribution
from .error import Error
from .error_details import ErrorDetails
from .evaluation import Evaluation
from .feature_config import FeatureConfig
from .feature_config_kind import FeatureConfigKind
from .feature_state import FeatureState
from .group_serving_rule import GroupServingRule
from .key_value import KeyValue
from .metrics import Metrics
from .metrics_data import MetricsData
from .metrics_data_metrics_type import MetricsDataMetricsType
from .pagination import Pagination
from .prerequisite import Prerequisite
from .proxy_config import ProxyConfig
from .proxy_config_environments_item import ProxyConfigEnvironmentsItem
from .segment import Segment
from .serve import Serve
from .serving_rule import ServingRule
from .tag import Tag
from .target import Target
from .target_attributes import TargetAttributes
from .target_data import TargetData
from .target_map import TargetMap
from .variation import Variation
from .variation_map import VariationMap
from .weighted_variation import WeightedVariation

__all__ = (
    "AuthenticateProxyKeyBody",
    "AuthenticationRequest",
    "AuthenticationRequestTarget",
    "AuthenticationRequestTargetAttributes",
    "AuthenticationResponse",
    "Clause",
    "Distribution",
    "Error",
    "ErrorDetails",
    "Evaluation",
    "FeatureConfig",
    "FeatureConfigKind",
    "FeatureState",
    "GroupServingRule",
    "KeyValue",
    "Metrics",
    "MetricsData",
    "MetricsDataMetricsType",
    "Pagination",
    "Prerequisite",
    "ProxyConfig",
    "ProxyConfigEnvironmentsItem",
    "Segment",
    "Serve",
    "ServingRule",
    "Tag",
    "Target",
    "TargetAttributes",
    "TargetData",
    "TargetMap",
    "Variation",
    "VariationMap",
    "WeightedVariation",
)



from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="WeightedVariation")


@_attrs_define
class WeightedVariation:
    """A variation and the weighting it should receive as part of a percentage rollout

    Attributes:
        variation (str): The variation identifier Example: off-variation.
        weight (int): The weight to be given to the variation in percent Example: 50.
    """

    variation: str
    weight: int
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        variation = self.variation

        weight = self.weight

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "variation": variation,
                "weight": weight,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        variation = d.pop("variation")

        weight = d.pop("weight")

        weighted_variation = cls(
            variation=variation,
            weight=weight,
        )

        weighted_variation.additional_properties = d
        return weighted_variation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FeatureConfigKind(str, Enum):
    BOOLEAN = "boolean"
    INT = "int"
    JSON = "json"
    STRING = "string"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.weighted_variation import WeightedVariation


T = TypeVar("T", bound="Distribution")


@_attrs_define
class Distribution:
    """Describes a distribution rule

    Attributes:
        bucket_by (str): The attribute to use when distributing targets across buckets
        variations (List['WeightedVariation']): A list of variations and the weight that should be given to each
    """

    bucket_by: str
    variations: List["WeightedVariation"]
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        bucket_by = self.bucket_by

        variations = []
        for variations_item_data in self.variations:
            variations_item = variations_item_data.to_dict()
            variations.append(variations_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "bucketBy": bucket_by,
                "variations": variations,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.weighted_variation import WeightedVariation

        d = src_dict.copy()
        bucket_by = d.pop("bucketBy")

        variations = []
        _variations = d.pop("variations")
        for variations_item_data in _variations:
            variations_item = WeightedVariation.from_dict(variations_item_data)

            variations.append(variations_item)

        distribution = cls(
            bucket_by=bucket_by,
            variations=variations,
        )

        distribution.additional_properties = d
        return distribution

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .error import Error
from .error_details import ErrorDetails
from .key_value import KeyValue
from .metrics import Metrics
from .metrics_data import MetricsData
from .metrics_data_metrics_type import MetricsDataMetricsType
from .target_data import TargetData

__all__ = (
    "Error",
    "ErrorDetails",
    "KeyValue",
    "Metrics",
    "MetricsData",
    "MetricsDataMetricsType",
    "TargetData",
)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.feature_config_kind import FeatureConfigKind
from ..models.feature_state import FeatureState
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.prerequisite import Prerequisite
    from ..models.serve import Serve
    from ..models.serving_rule import ServingRule
    from ..models.variation import Variation
    from ..models.variation_map import VariationMap


T = TypeVar("T", bound="FeatureConfig")


@_attrs_define
class FeatureConfig:
    """
    Attributes:
        project (str):
        environment (str):
        feature (str):
        state (FeatureState): The state of a flag either off or on
        kind (FeatureConfigKind):
        variations (List['Variation']):
        default_serve (Serve): Describe the distribution rule and the variation that should be served to the target
        off_variation (str):
        rules (Union[Unset, List['ServingRule']]):
        prerequisites (Union[Unset, List['Prerequisite']]):
        variation_to_target_map (Union[Unset, List['VariationMap']]):
        version (Union[Unset, int]):
    """

    project: str
    environment: str
    feature: str
    state: FeatureState
    kind: FeatureConfigKind
    variations: List["Variation"]
    default_serve: "Serve"
    off_variation: str
    rules: Union[Unset, List["ServingRule"]] = UNSET
    prerequisites: Union[Unset, List["Prerequisite"]] = UNSET
    variation_to_target_map: Union[Unset, List["VariationMap"]] = UNSET
    version: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        project = self.project

        environment = self.environment

        feature = self.feature

        state = self.state.value

        kind = self.kind.value

        variations = []
        for variations_item_data in self.variations:
            variations_item = variations_item_data.to_dict()
            variations.append(variations_item)

        default_serve = self.default_serve.to_dict()

        off_variation = self.off_variation

        rules: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.rules, Unset):
            rules = []
            for rules_item_data in self.rules:
                rules_item = rules_item_data.to_dict()
                rules.append(rules_item)

        prerequisites: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.prerequisites, Unset):
            prerequisites = []
            for prerequisites_item_data in self.prerequisites:
                prerequisites_item = prerequisites_item_data.to_dict()
                prerequisites.append(prerequisites_item)

        variation_to_target_map: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.variation_to_target_map, Unset):
            variation_to_target_map = []
            for variation_to_target_map_item_data in self.variation_to_target_map:
                variation_to_target_map_item = (
                    variation_to_target_map_item_data.to_dict()
                )
                variation_to_target_map.append(variation_to_target_map_item)

        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "project": project,
                "environment": environment,
                "feature": feature,
                "state": state,
                "kind": kind,
                "variations": variations,
                "defaultServe": default_serve,
                "offVariation": off_variation,
            }
        )
        if rules is not UNSET:
            field_dict["rules"] = rules
        if prerequisites is not UNSET:
            field_dict["prerequisites"] = prerequisites
        if variation_to_target_map is not UNSET:
            field_dict["variationToTargetMap"] = variation_to_target_map
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.prerequisite import Prerequisite
        from ..models.serve import Serve
        from ..models.serving_rule import ServingRule
        from ..models.variation import Variation
        from ..models.variation_map import VariationMap

        d = src_dict.copy()
        project = d.pop("project")

        environment = d.pop("environment")

        feature = d.pop("feature")

        state = FeatureState(d.pop("state"))

        kind = FeatureConfigKind(d.pop("kind"))

        variations = []
        _variations = d.pop("variations")
        for variations_item_data in _variations:
            variations_item = Variation.from_dict(variations_item_data)

            variations.append(variations_item)

        default_serve = Serve.from_dict(d.pop("defaultServe"))

        off_variation = d.pop("offVariation")

        rules = []
        _rules = d.pop("rules", UNSET)
        for rules_item_data in _rules or []:
            rules_item = ServingRule.from_dict(rules_item_data)

            rules.append(rules_item)

        prerequisites = []
        _prerequisites = d.pop("prerequisites", UNSET)
        for prerequisites_item_data in _prerequisites or []:
            prerequisites_item = Prerequisite.from_dict(
                prerequisites_item_data)

            prerequisites.append(prerequisites_item)

        variation_to_target_map = []
        _variation_to_target_map = d.pop("variationToTargetMap", UNSET)
        for variation_to_target_map_item_data in _variation_to_target_map or []:
            variation_to_target_map_item = VariationMap.from_dict(
                variation_to_target_map_item_data
            )

            variation_to_target_map.append(variation_to_target_map_item)

        version = d.pop("version", UNSET)

        feature_config = cls(
            project=project,
            environment=environment,
            feature=feature,
            state=state,
            kind=kind,
            variations=variations,
            default_serve=default_serve,
            off_variation=off_variation,
            rules=rules,
            prerequisites=prerequisites,
            variation_to_target_map=variation_to_target_map,
            version=version,
        )

        feature_config.additional_properties = d
        return feature_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.clause import Clause
    from ..models.serve import Serve


T = TypeVar("T", bound="ServingRule")


@_attrs_define
class ServingRule:
    """The rule used to determine what variation to serve to a target

    Attributes:
        priority (int): The rules priority relative to other rules.  The rules are evaluated in order with 1 being the
            highest Example: 1.
        clauses (List['Clause']): A list of clauses to use in the rule
        serve (Serve): Describe the distribution rule and the variation that should be served to the target
        rule_id (Union[Unset, str]): The unique identifier for this rule
    """

    priority: int
    clauses: List["Clause"]
    serve: "Serve"
    rule_id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        priority = self.priority

        clauses = []
        for clauses_item_data in self.clauses:
            clauses_item = clauses_item_data.to_dict()
            clauses.append(clauses_item)

        serve = self.serve.to_dict()

        rule_id = self.rule_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "priority": priority,
                "clauses": clauses,
                "serve": serve,
            }
        )
        if rule_id is not UNSET:
            field_dict["ruleId"] = rule_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.clause import Clause
        from ..models.serve import Serve

        d = src_dict.copy()
        priority = d.pop("priority")

        clauses = []
        _clauses = d.pop("clauses")
        for clauses_item_data in _clauses:
            clauses_item = Clause.from_dict(clauses_item_data)

            clauses.append(clauses_item)

        serve = Serve.from_dict(d.pop("serve"))

        rule_id = d.pop("ruleId", UNSET)

        serving_rule = cls(
            priority=priority,
            clauses=clauses,
            serve=serve,
            rule_id=rule_id,
        )

        serving_rule.additional_properties = d
        return serving_rule

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.segment import Segment
    from ..models.target_attributes import TargetAttributes


T = TypeVar("T", bound="Target")


@_attrs_define
class Target:
    """A Target object

    Attributes:
        identifier (str): The unique identifier for this target Example: john-doe.
        account (str): The account ID that the target belongs to Example: abcXDdffdaffd.
        org (str): The identifier for the organization that the target belongs to
        environment (str): The identifier for the environment that the target belongs to
        project (str): The identifier for the project that this target belongs to
        name (str): The name of this Target Example: John Doe.
        anonymous (Union[Unset, bool]): Indicates if this target is anonymous
        attributes (Union[Unset, TargetAttributes]): a JSON representation of the attributes for this target Example:
            {'age': 20, 'location': 'Belfast'}.
        created_at (Union[Unset, int]): The date and time in milliseconds when this Target was created
        segments (Union[Unset, List['Segment']]): A list of Target Groups (Segments) that this Target belongs to
    """

    identifier: str
    account: str
    org: str
    environment: str
    project: str
    name: str
    anonymous: Union[Unset, bool] = UNSET
    attributes: Union[Unset, "TargetAttributes"] = UNSET
    created_at: Union[Unset, int] = UNSET
    segments: Union[Unset, List["Segment"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier

        account = self.account

        org = self.org

        environment = self.environment

        project = self.project

        name = self.name

        anonymous = self.anonymous

        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        created_at = self.created_at

        segments: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.segments, Unset):
            segments = []
            for segments_item_data in self.segments:
                segments_item = segments_item_data.to_dict()
                segments.append(segments_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "account": account,
                "org": org,
                "environment": environment,
                "project": project,
                "name": name,
            }
        )
        if anonymous is not UNSET:
            field_dict["anonymous"] = anonymous
        if attributes is not UNSET:
            field_dict["attributes"] = attributes
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if segments is not UNSET:
            field_dict["segments"] = segments

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.segment import Segment
        from ..models.target_attributes import TargetAttributes

        d = src_dict.copy()
        identifier = d.pop("identifier")

        account = d.pop("account")

        org = d.pop("org")

        environment = d.pop("environment")

        project = d.pop("project")

        name = d.pop("name")

        anonymous = d.pop("anonymous", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, TargetAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = TargetAttributes.from_dict(_attributes)

        created_at = d.pop("createdAt", UNSET)

        segments = []
        _segments = d.pop("segments", UNSET)
        for segments_item_data in _segments or []:
            segments_item = Segment.from_dict(segments_item_data)

            segments.append(segments_item)

        target = cls(
            identifier=identifier,
            account=account,
            org=org,
            environment=environment,
            project=project,
            name=name,
            anonymous=anonymous,
            attributes=attributes,
            created_at=created_at,
            segments=segments,
        )

        target.additional_properties = d
        return target

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.feature_config import FeatureConfig
    from ..models.segment import Segment


T = TypeVar("T", bound="ProxyConfigEnvironmentsItem")


@_attrs_define
class ProxyConfigEnvironmentsItem:
    """
    Attributes:
        id (Union[Unset, str]):
        api_keys (Union[Unset, List[str]]):
        feature_configs (Union[Unset, List['FeatureConfig']]):
        segments (Union[Unset, List['Segment']]):
    """

    id: Union[Unset, str] = UNSET
    api_keys: Union[Unset, List[str]] = UNSET
    feature_configs: Union[Unset, List["FeatureConfig"]] = UNSET
    segments: Union[Unset, List["Segment"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        api_keys: Union[Unset, List[str]] = UNSET
        if not isinstance(self.api_keys, Unset):
            api_keys = self.api_keys

        feature_configs: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.feature_configs, Unset):
            feature_configs = []
            for feature_configs_item_data in self.feature_configs:
                feature_configs_item = feature_configs_item_data.to_dict()
                feature_configs.append(feature_configs_item)

        segments: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.segments, Unset):
            segments = []
            for segments_item_data in self.segments:
                segments_item = segments_item_data.to_dict()
                segments.append(segments_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if api_keys is not UNSET:
            field_dict["apiKeys"] = api_keys
        if feature_configs is not UNSET:
            field_dict["featureConfigs"] = feature_configs
        if segments is not UNSET:
            field_dict["segments"] = segments

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.feature_config import FeatureConfig
        from ..models.segment import Segment

        d = src_dict.copy()
        id = d.pop("id", UNSET)

        api_keys = cast(List[str], d.pop("apiKeys", UNSET))

        feature_configs = []
        _feature_configs = d.pop("featureConfigs", UNSET)
        for feature_configs_item_data in _feature_configs or []:
            feature_configs_item = FeatureConfig.from_dict(
                feature_configs_item_data)

            feature_configs.append(feature_configs_item)

        segments = []
        _segments = d.pop("segments", UNSET)
        for segments_item_data in _segments or []:
            segments_item = Segment.from_dict(segments_item_data)

            segments.append(segments_item)

        proxy_config_environments_item = cls(
            id=id,
            api_keys=api_keys,
            feature_configs=feature_configs,
            segments=segments,
        )

        proxy_config_environments_item.additional_properties = d
        return proxy_config_environments_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.error_details import ErrorDetails


T = TypeVar("T", bound="Error")


@_attrs_define
class Error:
    """
    Attributes:
        code (str): The http error code Example: 404.
        message (str): The reason the request failed Example: Error retrieving projects, organization 'default_org' does
            not exist.
        details (Union[Unset, ErrorDetails]): Additional details about the error
    """

    code: str
    message: str
    details: Union[Unset, "ErrorDetails"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        code = self.code

        message = self.message

        details: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.details, Unset):
            details = self.details.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "code": code,
                "message": message,
            }
        )
        if details is not UNSET:
            field_dict["details"] = details

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.error_details import ErrorDetails

        d = src_dict.copy()
        code = d.pop("code")

        message = d.pop("message")

        _details = d.pop("details", UNSET)
        details: Union[Unset, ErrorDetails]
        if isinstance(_details, Unset):
            details = UNSET
        else:
            details = ErrorDetails.from_dict(_details)

        error = cls(
            code=code,
            message=message,
            details=details,
        )

        error.additional_properties = d
        return error

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Variation")


@_attrs_define
class Variation:
    """A variation of a flag that can be returned to a target

    Attributes:
        identifier (str): The unique identifier for the variation Example: off-variation.
        value (str): The variation value to serve such as true or false for a boolean flag Example: true.
        name (Union[Unset, str]): The user friendly name of the variation Example: Off VAriation.
        description (Union[Unset, str]): A description of the variation
    """

    identifier: str
    value: str
    name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier

        value = self.value

        name = self.name

        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "value": value,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        value = d.pop("value")

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        variation = cls(
            identifier=identifier,
            value=value,
            name=name,
            description=description,
        )

        variation.additional_properties = d
        return variation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.proxy_config_environments_item import \
        ProxyConfigEnvironmentsItem


T = TypeVar("T", bound="ProxyConfig")


@_attrs_define
class ProxyConfig:
    """TBD

    Attributes:
        page_count (int): The total number of pages Example: 100.
        item_count (int): The total number of items Example: 1.
        page_size (int): The number of items per page Example: 1.
        page_index (int): The current page
        version (Union[Unset, int]): The version of this object.  The version will be incremented each time the object
            is modified Example: 5.
        environments (Union[Unset, List['ProxyConfigEnvironmentsItem']]):
    """

    page_count: int
    item_count: int
    page_size: int
    page_index: int
    version: Union[Unset, int] = UNSET
    environments: Union[Unset, List["ProxyConfigEnvironmentsItem"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        page_count = self.page_count

        item_count = self.item_count

        page_size = self.page_size

        page_index = self.page_index

        version = self.version

        environments: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.environments, Unset):
            environments = []
            for environments_item_data in self.environments:
                environments_item = environments_item_data.to_dict()
                environments.append(environments_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "pageCount": page_count,
                "itemCount": item_count,
                "pageSize": page_size,
                "pageIndex": page_index,
            }
        )
        if version is not UNSET:
            field_dict["version"] = version
        if environments is not UNSET:
            field_dict["environments"] = environments

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.proxy_config_environments_item import \
            ProxyConfigEnvironmentsItem

        d = src_dict.copy()
        page_count = d.pop("pageCount")

        item_count = d.pop("itemCount")

        page_size = d.pop("pageSize")

        page_index = d.pop("pageIndex")

        version = d.pop("version", UNSET)

        environments = []
        _environments = d.pop("environments", UNSET)
        for environments_item_data in _environments or []:
            environments_item = ProxyConfigEnvironmentsItem.from_dict(
                environments_item_data
            )

            environments.append(environments_item)

        proxy_config = cls(
            page_count=page_count,
            item_count=item_count,
            page_size=page_size,
            page_index=page_index,
            version=version,
            environments=environments,
        )

        proxy_config.additional_properties = d
        return proxy_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Client for interacting with Harness FF server"""

import json
import threading
from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

from jwt import decode

import featureflags.sdk_logging_codes as sdk_codes
from featureflags.analytics import AnalyticsService
from featureflags.evaluations.evaluator import (Evaluator,
                                                FlagKindMismatchException)
from featureflags.repository import Repository

from .api import UnrecoverableRequestException, retryable_authenticate
from .config import Config, default_config
from .evaluations.auth_target import Target
from .openapi.config.api.client.authenticate import AuthenticationRequest
from .openapi.config.client import AuthenticatedClient, Client
from .polling import PollingProcessor
from .streaming import StreamProcessor
from .util import log

VERSION: str = "1.6.1"


class MissingOrEmptyAPIKeyException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"MissingOrEmptyAPIKeyException: {self.message}"


class FeatureFlagType(str, Enum):
    BOOLEAN = "boolean"
    INT_OR_FLOAT = "int"
    STRING = "string"
    JSON = "json"
    FLAG_NOT_FOUND = "flag_not_found"


class CfClient(object):
    def __init__(
            self, sdk_key: str,
            *options: Callable,
            config: Optional[Config] = None
    ):
        self._client: Optional[Client] = None
        #  The Client is considered initialized when flags and groups
        #  are loaded into cache.
        self._initialized = threading.Event()
        # Keep track if initialization has failed due to authentication
        # or a missing/empty API key.
        self._initialized_failed = False
        self._initialised_failed_reason: Dict[bool, Optional[str]] \
            = {False: None, True: None}
        self._auth_token: Optional[str] = None
        self._environment_id: Optional[str] = None
        self._sdk_key: Optional[str] = sdk_key
        self._config: Config = default_config
        self._cluster: str = '1'
        self._account_id = None

        if config:
            self._config = config

        for option in options:
            if callable(option):
                option(self._config)

        if self._config.cache is None:
            raise Exception("cache cannot be none")

        self._repository = Repository(self._config.cache)
        self._evaluator = Evaluator(self._repository)

        self.run()

    def run(self):
        try:
            if self._sdk_key is None:
                raise MissingOrEmptyAPIKeyException("SDK Key is None")

            if self._sdk_key == "":
                raise MissingOrEmptyAPIKeyException("SDK Key is an empty "
                                                    "string")

            self.authenticate()
            sdk_codes.info_sdk_auth_ok()
            streaming_event = threading.Event()
            polling_event = threading.Event()

            self._polling_processor = PollingProcessor(
                client=self._client,
                config=self._config,
                environment_id=self._environment_id,
                #  PollingProcessor is responsible for doing the initial
                #  flag/group fetch and cache. So we allocate it the
                #  responsibility
                #  for setting the Client is_initialized variable.
                wait_for_initialization=self._initialized,
                initialised_failed_reason=self._initialised_failed_reason,
                ready=polling_event,
                stream_ready=streaming_event,
                repository=self._repository,
                cluster=self._cluster,
            )
            self._polling_processor.start()

            if self._config.enable_stream:
                self._stream = StreamProcessor(
                    repository=self._repository,
                    client=self._client,
                    environment_id=self._environment_id,
                    api_key=self._sdk_key,
                    token=self._auth_token,
                    config=self._config,
                    ready=streaming_event,
                    poller=polling_event,
                    cluster=self._cluster,
                )
                self._stream.start()

            if self._config.enable_analytics:
                metrics_client = self.make_client(
                    self._config.events_url,
                    self._auth_token,
                    self._account_id,
                    config=self._config)
                self._analytics = AnalyticsService(
                    config=self._config,
                    client=metrics_client,
                    environment=self._environment_id,
                    cluster=self._cluster,
                )

        except UnrecoverableRequestException as e:

            sdk_codes.warn_auth_failed_exceed_retries()
            sdk_codes.warn_failed_init_auth_error(e)
            self._initialized_failed = True
            # We just need to unblock the thread here
            # in case wait_for_intialization was called. The SDK has already
            # logged that authentication failed and defaults will be returned.
            self._initialized.set()
        except MissingOrEmptyAPIKeyException:
            self._initialized_failed = True
            self._initialised_failed_reason[True] \
                = str(MissingOrEmptyAPIKeyException)
            sdk_codes.wan_missing_sdk_key()
            # And again, unblock the thread.
            self._initialized.set()
        except Exception as ex:
            sdk_codes.warn_failed_init_auth_error(ex)
            self._initialised_failed_reason[True] = str(ex)
            self._initialized.set()

    def wait_for_initialization(self):
        sdk_codes.info_sdk_init_waiting()
        self._initialized.wait()

    def is_initialized(self):
        if self._initialized_failed \
                or self._initialised_failed_reason[True] is not None:
            return False
        return self._initialized.is_set()

    def get_environment_id(self):
        return self._environment_id

    def authenticate(self):
        verify = True
        if self._config.tls_trusted_cas_file is not None:
            verify = self._config.tls_trusted_cas_file

        client = Client(base_url=self._config.base_url, verify_ssl=verify,
                        raise_on_unexpected_status=True)
        body = AuthenticationRequest(api_key=self._sdk_key)
        response = retryable_authenticate(client=client, body=body).parsed
        self._auth_token = response.auth_token

        decoded = decode(self._auth_token, options={
            "verify_signature": False})
        self._environment_id = decoded["environment"]
        self._cluster = decoded["clusterIdentifier"]
        if not self._cluster:
            self._cluster = '1'
        if "accountID" in decoded:
            self._account_id = decoded["accountID"]

        self._client = self.make_client(
            self._config.base_url,
            self._auth_token,
            self._account_id,
            config=self._config)

    def make_client(self, url, token, account_id, config):

        verify = True
        if config.tls_trusted_cas_file is not None:
            verify = self._config.tls_trusted_cas_file

        client = AuthenticatedClient(
            base_url=url,
            token=token,
            verify_ssl=verify
        )
        # Additional headers used to track usage
        additional_headers = {
            "User-Agent": "PythonSDK/" + VERSION,
            "Harness-SDK-Info": f'Python {VERSION} Server',
            "Harness-EnvironmentID": self._environment_id}
        # At present the FF Relay Proxy does not send the accountID claim
        if account_id:
            additional_headers["Harness-AccountID"] = account_id
        client.with_headers(additional_headers)
        return client

    def get_flag_type(self, identifier) -> Optional[FeatureFlagType]:
        if self._initialised_failed_reason[True] is not None:
            log.warning(
                "Failed to check flag type for flag '%s', reason: Client is "
                "not initialized",
                identifier)
            return FeatureFlagType("flag_not_found")
        kind = self._evaluator.get_kind(identifier)
        if not kind:
            log.warning(
                "Failed to check flag kind for flag '%s', reason: flag not "
                "found", identifier)
            return FeatureFlagType("flag_not_found")
        return FeatureFlagType(kind)

    def bool_variation(self, identifier: str, target: Target,
                       default: bool) -> bool:
        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate bool variation for flag '%s'"
                " and the default variation '%s' is being returned. Reason: "
                "`Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(identifier, target, "boolean")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s "
                    "and the "
                    "default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)
            return variation.value.lower() == "true"

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate bool variation for flag '%s'"
                " and the default variation '%s' is being returned. Reason: "
                "'%s'", identifier, default, str(ex))
            return default

    def int_variation(self, identifier: str, target: Target,
                      default: int) -> int:

        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate bool variation for flag '%s'"
                " and the default variation '%s' is being returned. Reason: "
                "`Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(identifier, target, "int")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s "
                    "and the "
                    "default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)

            return int(variation.value)

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate int variation for flag '%s'"
                " and the default variation '%s' is being returned. Reason: "
                "'%s'", identifier, default, str(ex))
            return default

    def number_variation(self, identifier: str, target: Target,
                         default: float) -> float:

        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate number variation for flag "
                "'%s' and the default variation '%s' is being returned. "
                "Reason: `Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(
                identifier, target, "int")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s "
                    "and the "
                    "default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)

            return float(variation.value)

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate number variation for flag "
                "'%s'  and the default variation '%s' is being returned. "
                "Reason: '%s'", identifier, default, str(ex))
            return default

    def int_or_float_variation(self, identifier: str, target: Target,
                               default: Union[float, int]) -> \
            Union[float, int]:

        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate int_or_float variation for "
                "flag '%s' and the default variation '%s' is being returned. "
                "Reason: `Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(
                identifier, target, "int")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s "
                    "and the "
                    "default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)

            try:
                result = int(variation.value)
            except ValueError:
                try:
                    result = float(variation.value)
                except ValueError:
                    # If both conversions fail, log an error and return the
                    # default
                    log.error(
                        "SDKCODE:6001: Invalid number format for %s. "
                        "Expected a number but got '%s'",
                        {"flag": identifier, "value": variation.value}
                    )
                    return default

            return result

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate int_or_float variation for "
                "flag '%s'  and the default variation '%s' is being returned. "
                "Reason: '%s'", identifier, default, str(ex))
            return default

    def string_variation(self, identifier: str, target: Target,
                         default: str) -> str:

        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate string variation for flag "
                "'%s' and the default variation '%s' is being returned. "
                "Reason: `Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(
                identifier, target, "string")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s "
                    "and the "
                    "default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)

            return variation.value

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate string variation for flag "
                "'%s' and the default variation '%s' is being returned. "
                "Reason: '%s'", identifier, default, str(ex))
            return default

    def json_variation(self, identifier: str, target: Target,
                       default: Dict[str, Any]) -> Dict[str, Any]:

        # If initialization has failed, then return the default variation
        # immediately
        if self._initialised_failed_reason[True] is not None:
            log.error(
                "SDKCODE:6001: Failed to evaluate json variation for flag '%s'"
                " and the default variation '%s' is being returned. Reason: "
                "`Client is not initialized: %s'",
                identifier, default, self._initialised_failed_reason[True])
            return default

        try:
            variation = self._evaluator.evaluate(identifier, target, "json")

            if not variation or not variation.value:
                log.error(
                    "SDKCODE:6001: Failed to evaluate bool variation for %s"
                    " and the default variation '%s' is being returned",
                    {"target": target, "flag": identifier}, default)
                return default

            # Only register metrics if analytics is enabled,
            # and sometimes when the SDK starts up we can
            # evaluate before the flag is cached which results in
            # an empty identifier.
            if self._config.enable_analytics and variation.identifier != "":
                self._analytics.enqueue(target, identifier, variation)

            return json.loads(variation.value)

        except FlagKindMismatchException as ex:
            log.error(
                "SDKCODE:6001: Failed to evaluate json variation for flag "
                "'%s' and the default variation '%s' is being returned. "
                "Reason: '%s'", identifier, default, str(ex))
            return default

    def close(self):
        sdk_codes.info_sdk_start_close()
        self._polling_processor.stop()
        if self._config.enable_stream:
            self._stream.stop()

        if self._config.enable_analytics:
            self._analytics.close()
        sdk_codes.info_sdk_close_success()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Evaluation")


@_attrs_define
class Evaluation:
    """
    Attributes:
        flag (str):
        value (str):
        kind (str):
        identifier (Union[Unset, str]):
    """

    flag: str
    value: str
    kind: str
    identifier: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        flag = self.flag

        value = self.value

        kind = self.kind

        identifier = self.identifier

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "flag": flag,
                "value": value,
                "kind": kind,
            }
        )
        if identifier is not UNSET:
            field_dict["identifier"] = identifier

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        flag = d.pop("flag")

        value = d.pop("value")

        kind = d.pop("kind")

        identifier = d.pop("identifier", UNSET)

        evaluation = cls(
            flag=flag,
            value=value,
            kind=kind,
            identifier=identifier,
        )

        evaluation.additional_properties = d
        return evaluation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.clause import Clause
    from ..models.group_serving_rule import GroupServingRule
    from ..models.tag import Tag
    from ..models.target import Target


T = TypeVar("T", bound="Segment")


@_attrs_define
class Segment:
    """A Target Group (Segment) response

    Attributes:
        identifier (str): Unique identifier for the target group.
        name (str): Name of the target group. Example: Beta Testers.
        environment (Union[Unset, str]): The environment this target group belongs to Example: Production.
        tags (Union[Unset, List['Tag']]): Tags for this target group
        included (Union[Unset, List['Target']]): A list of Targets who belong to this target group
        excluded (Union[Unset, List['Target']]): A list of Targets who are excluded from this target group
        rules (Union[Unset, List['Clause']]):
        serving_rules (Union[Unset, List['GroupServingRule']]): An array of rules that can cause a user to be included
            in this segment.
        created_at (Union[Unset, int]): The data and time in milliseconds when the group was created
        modified_at (Union[Unset, int]): The data and time in milliseconds when the group was last modified
        version (Union[Unset, int]): The version of this group.  Each time it is modified the version is incremented
            Example: 1.
    """

    identifier: str
    name: str
    environment: Union[Unset, str] = UNSET
    tags: Union[Unset, List["Tag"]] = UNSET
    included: Union[Unset, List["Target"]] = UNSET
    excluded: Union[Unset, List["Target"]] = UNSET
    rules: Union[Unset, List["Clause"]] = UNSET
    serving_rules: Union[Unset, List["GroupServingRule"]] = UNSET
    created_at: Union[Unset, int] = UNSET
    modified_at: Union[Unset, int] = UNSET
    version: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier

        name = self.name

        environment = self.environment

        tags: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = []
            for tags_item_data in self.tags:
                tags_item = tags_item_data.to_dict()
                tags.append(tags_item)

        included: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.included, Unset):
            included = []
            for included_item_data in self.included:
                included_item = included_item_data.to_dict()
                included.append(included_item)

        excluded: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.excluded, Unset):
            excluded = []
            for excluded_item_data in self.excluded:
                excluded_item = excluded_item_data.to_dict()
                excluded.append(excluded_item)

        rules: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.rules, Unset):
            rules = []
            for rules_item_data in self.rules:
                rules_item = rules_item_data.to_dict()
                rules.append(rules_item)

        serving_rules: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.serving_rules, Unset):
            serving_rules = []
            for serving_rules_item_data in self.serving_rules:
                serving_rules_item = serving_rules_item_data.to_dict()
                serving_rules.append(serving_rules_item)

        created_at = self.created_at

        modified_at = self.modified_at

        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "name": name,
            }
        )
        if environment is not UNSET:
            field_dict["environment"] = environment
        if tags is not UNSET:
            field_dict["tags"] = tags
        if included is not UNSET:
            field_dict["included"] = included
        if excluded is not UNSET:
            field_dict["excluded"] = excluded
        if rules is not UNSET:
            field_dict["rules"] = rules
        if serving_rules is not UNSET:
            field_dict["servingRules"] = serving_rules
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if modified_at is not UNSET:
            field_dict["modifiedAt"] = modified_at
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.clause import Clause
        from ..models.group_serving_rule import GroupServingRule
        from ..models.tag import Tag
        from ..models.target import Target

        d = src_dict.copy()
        identifier = d.pop("identifier")

        name = d.pop("name")

        environment = d.pop("environment", UNSET)

        tags = []
        _tags = d.pop("tags", UNSET)
        for tags_item_data in _tags or []:
            tags_item = Tag.from_dict(tags_item_data)

            tags.append(tags_item)

        included = []
        _included = d.pop("included", UNSET)
        for included_item_data in _included or []:
            included_item = Target.from_dict(included_item_data)

            included.append(included_item)

        excluded = []
        _excluded = d.pop("excluded", UNSET)
        for excluded_item_data in _excluded or []:
            excluded_item = Target.from_dict(excluded_item_data)

            excluded.append(excluded_item)

        rules = []
        _rules = d.pop("rules", UNSET)
        for rules_item_data in _rules or []:
            rules_item = Clause.from_dict(rules_item_data)

            rules.append(rules_item)

        serving_rules = []
        _serving_rules = d.pop("servingRules", UNSET)
        for serving_rules_item_data in _serving_rules or []:
            serving_rules_item = GroupServingRule.from_dict(
                serving_rules_item_data)

            serving_rules.append(serving_rules_item)

        created_at = d.pop("createdAt", UNSET)

        modified_at = d.pop("modifiedAt", UNSET)

        version = d.pop("version", UNSET)

        segment = cls(
            identifier=identifier,
            name=name,
            environment=environment,
            tags=tags,
            included=included,
            excluded=excluded,
            rules=rules,
            serving_rules=serving_rules,
            created_at=created_at,
            modified_at=modified_at,
            version=version,
        )

        segment.additional_properties = d
        return segment

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Clause")


@_attrs_define
class Clause:
    """A clause describes what conditions are used to evaluate a flag

    Attributes:
        attribute (str): The attribute to use in the clause.  This can be any target attribute Example: identifier.
        op (str): The type of operation such as equals, starts_with, contains Example: starts_with.
        values (List[str]): The values that are compared against the operator
        negate (bool): Is the operation negated?
        id (Union[Unset, str]): The unique ID for the clause Example: 32434243.
    """

    attribute: str
    op: str
    values: List[str]
    negate: bool
    id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        attribute = self.attribute

        op = self.op

        values = self.values

        negate = self.negate

        id = self.id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "attribute": attribute,
                "op": op,
                "values": values,
                "negate": negate,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        attribute = d.pop("attribute")

        op = d.pop("op")

        values = cast(List[str], d.pop("values"))

        negate = d.pop("negate")

        id = d.pop("id", UNSET)

        clause = cls(
            attribute=attribute,
            op=op,
            values=values,
            negate=negate,
            id=id,
        )

        clause.additional_properties = d
        return clause

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import concurrent.futures
import time
import traceback
from threading import Lock, Thread
from typing import Dict, List, Set, Union

import attr
import httpx

from .config import Config
from .openapi.config import AuthenticatedClient
from .openapi.config.models.target import Target
from .openapi.config.models.target_attributes import TargetAttributes
from .openapi.config.models.variation import Variation
from .openapi.metrics.api.metrics.post_metrics import \
    sync_detailed as post_metrics
from .openapi.metrics.models.key_value import KeyValue
from .openapi.metrics.models.metrics import Metrics
from .openapi.metrics.models.metrics_data import MetricsData
from .openapi.metrics.models.metrics_data_metrics_type import \
    MetricsDataMetricsType
from .openapi.metrics.models.target_data import TargetData
from .openapi.metrics.types import Unset
from .sdk_logging_codes import (info_evaluation_metrics_exceeded,
                                info_metrics_success,
                                info_metrics_target_batch_success,
                                info_metrics_target_exceeded,
                                info_metrics_thread_existed,
                                info_metrics_thread_started,
                                warn_post_metrics_failed,
                                warn_post_metrics_target_batch_failed)
from .util import log

FF_METRIC_TYPE = 'FFMETRICS'
FEATURE_IDENTIFIER_ATTRIBUTE = 'featureIdentifier'
FEATURE_NAME_ATTRIBUTE = 'featureName'
VARIATION_IDENTIFIER_ATTRIBUTE = 'variationIdentifier'
VARIATION_VALUE_ATTRIBUTE = 'variationValue'
TARGET_ATTRIBUTE = 'target'
SDK_VERSION_ATTRIBUTE = 'SDK_VERSION'
SDK_VERSION = '1.6.1'
SDK_TYPE_ATTRIBUTE = 'SDK_TYPE'
SDK_TYPE = 'server'
SDK_LANGUAGE_ATTRIBUTE = 'SDK_LANGUAGE'
SDK_LANGUAGE = 'python'
GLOBAL_TARGET = 'global'


@attr.s(auto_attribs=True)
class AnalyticsEvent(object):
    target: Target
    flag_identifier: str
    variation: Variation
    count: int = 0


@attr.s(auto_attribs=True)
class MetricTargetData(object):
    identifier: str
    name: str
    attributes: Union[Unset, TargetAttributes] = Unset


class AnalyticsService(object):

    def __init__(self, config: Config,
                 client: AuthenticatedClient,
                 environment: str,
                 cluster: str) -> None:
        self._lock = Lock()
        self._config = config
        self._client = client
        self._environment = environment
        self._cluster = cluster

        # Evaluation metrics
        self._data: Dict[str, AnalyticsEvent] = {}
        # This allows for up to 2K flags with 5 variations each per interval
        self._max_evaluation_metrics = 10000
        self._max_evaluation_metrics_exceeded = False

        # Target metrics - batch based
        self._target_data_batches: List[Dict[str, MetricTargetData]] = [{}]

        # This allows for 100k unique targets per interval
        self._max_number_of_target_batches = 100
        self._max_target_batch_size = 1000
        self._current_target_batch_index = 0
        self.max_target_data_exceeded = False

        # Allow us to track targets for the life of the SDK instance, to
        # prevent sending the same target more than once. This also helps
        # unique targets get processed each interval, because duplicate
        # targets are not making up the 100K limit.
        self._seen_targets: Set[str] = set()

        self._running = False
        self._runner = Thread(target=self._sync)
        self._runner.daemon = True
        self._runner.start()

    def enqueue(self, target: Target, identifier: str,
                variation: Variation):
        event: AnalyticsEvent = AnalyticsEvent(
            target=target,
            flag_identifier=identifier,
            variation=variation
        )

        self._lock.acquire()
        try:
            # Check if adding a new metric would exceed the 10,000 limit
            if len(self._data) < self._max_evaluation_metrics:

                # Store unique evaluation events. We map a unique evaluation
                # event to its count.
                unique_evaluation_key = self.get_key(event)
                if unique_evaluation_key in self._data:
                    self._data[unique_evaluation_key].count += 1
                else:
                    event.count = 1
                    self._data[unique_evaluation_key] = event
            else:
                if not self._max_evaluation_metrics_exceeded:
                    self._max_evaluation_metrics_exceeded = True
                    info_evaluation_metrics_exceeded()

            # Don't store this target if it is anonymous. Note, we currently
            # don't do this check above for evaluation metrics, because we use
            # the global target.
            if target.anonymous:
                return

            unique_target_key = self.get_target_key(event)

            # If we've seen this target before, don't process it
            if unique_target_key in self._seen_targets:
                return

            self._seen_targets.add(unique_target_key)

            # Check if we're on our final target batch - if we are, and we've
            # exceeded the max batch size just return early.
            if len(self._target_data_batches) >= \
                    self._max_number_of_target_batches:
                if len(self._target_data_batches[
                    self._current_target_batch_index]) >= \
                        self._max_target_batch_size:
                    if not self.max_target_data_exceeded:
                        self.max_target_data_exceeded = True
                        info_metrics_target_exceeded()
                    return

            if event.target is not None and not event.target.anonymous:

                # Store unique targets. If the target already exists
                # in any of the batches, don't continue processing it
                for batch in self._target_data_batches:
                    if unique_target_key in batch:
                        return

                # If we've exceeded the max batch size for the current
                # batch, then create a new batch and start using it.
                if len(self._target_data_batches[
                    self._current_target_batch_index]) >= \
                        self._max_target_batch_size:
                    self._target_data_batches.append({})
                    self._current_target_batch_index += 1

                target_name = event.target.name
                # If the target has no name use the identifier
                if not target_name:
                    target_name = event.target.identifier
                self._target_data_batches[
                    self._current_target_batch_index][unique_target_key] = \
                    MetricTargetData(
                        identifier=event.target.identifier,
                        name=target_name,
                        attributes=event.target.attributes
                )

        finally:
            self._lock.release()

    # Returns a key for unique evaluations events.
    def get_key(self, event: AnalyticsEvent) -> str:
        return '{feature}-{variation}-{value}-{target}'.format(
            feature=event.flag_identifier,
            variation=event.variation.identifier,
            value=event.variation.value,
            target=GLOBAL_TARGET,
        )

    # Returns a key for unique targets. Targets are considered unique
    # if they have different identifiers.
    def get_target_key(self, event: AnalyticsEvent) -> str:
        return event.target.identifier

    def _sync(self) -> None:
        if not self._running:
            info_metrics_thread_started(
                f'{self._config.events_sync_interval}s')
            self._running = True
            while self._running:
                time.sleep(self._config.events_sync_interval)
                self._send_data()

    def _send_data(self) -> None:
        if not bool(self._data):
            log.debug('No metrics data!')
            return
        log.debug('Start sending metrics data')
        self._lock.acquire()
        target_data: List[TargetData] = []
        metrics_data: List[MetricsData] = []
        try:
            for _, event in self._data.items():
                metric_attributes: List[KeyValue] = [
                    KeyValue(FEATURE_IDENTIFIER_ATTRIBUTE,
                             event.flag_identifier),
                    KeyValue(FEATURE_NAME_ATTRIBUTE,
                             event.flag_identifier),
                    KeyValue(VARIATION_IDENTIFIER_ATTRIBUTE,
                             event.variation.identifier),
                    KeyValue(VARIATION_VALUE_ATTRIBUTE, event.variation.value),
                    KeyValue(SDK_TYPE_ATTRIBUTE, SDK_TYPE),
                    KeyValue(SDK_LANGUAGE_ATTRIBUTE, SDK_LANGUAGE),
                    KeyValue(SDK_VERSION_ATTRIBUTE, SDK_VERSION),
                    KeyValue(TARGET_ATTRIBUTE, GLOBAL_TARGET)
                ]

                md = MetricsData(
                    timestamp=int(round(time.time() * 1000)),
                    count=event.count,
                    metrics_type=MetricsDataMetricsType.FFMETRICS,
                    attributes=metric_attributes
                )
                metrics_data.append(md)
            for _, unique_target in self._target_data_batches[0].items():
                self.process_target(target_data, unique_target)

            target_data_batches: List[List[TargetData]] = []
            target_data_batch_index = 0
            # We've already accounted for the first batch, so start processing
            # from the second batch onwards
            for batch in self._target_data_batches[1:]:
                target_data_batches.append([])
                for _, unique_target in batch.items():
                    self.process_target(
                        target_data_batches[target_data_batch_index],
                        unique_target)
                target_data_batch_index += 1

        finally:
            self._data = {}
            self._target_data_batches = [{}]
            self._current_target_batch_index = 0
            self.max_target_data_exceeded = False
            self._max_evaluation_metrics_exceeded = False
            self._lock.release()

        body: Metrics = Metrics(target_data=target_data,
                                metrics_data=metrics_data)
        try:
            response = post_metrics(client=self._client,
                                    environment_uuid=self._environment,
                                    body=body,
                                    cluster=self._cluster)

            log.debug('Metrics server returns: %d', response.status_code)
            if response.status_code >= 400:
                warn_post_metrics_failed(response.status_code)
                return
            if len(target_data_batches) > 0:
                log.info('Sending %s target batches to metrics',
                         len(target_data_batches))
                unique_responses_codes = {}

                # Process batches concurrently
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    for batch in target_data_batches:
                        # Staggering requests over 0.02 seconds mean that we
                        # will send 200 requests every four seconds, so that
                        # the backend isn't hit too hard.
                        time.sleep(0.02)
                        future = executor.submit(
                            self.process_target_data_batch,
                            batch)
                        futures.append(future)

                    # Wait for all batches to complete
                    concurrent.futures.wait(futures)

                    # Get unique status codes
                    for future in futures:
                        status_code = future.result()
                        if status_code in unique_responses_codes:
                            unique_responses_codes[status_code] += 1
                        else:
                            unique_responses_codes[status_code] = 1

                # Log any error codes
                for unique_code, count in unique_responses_codes.items():
                    if unique_code >= 400:
                        warn_post_metrics_target_batch_failed(
                            f'{count} batches received code {unique_code}')
                        continue
                    info_metrics_target_batch_success(
                        f'{count} batches successful')

            info_metrics_success()
        except httpx.RequestError as ex:
            print(traceback.format_exc())
            warn_post_metrics_failed(ex)

    def process_target_data_batch(self, target_data_batch):
        batch_request_body: Metrics = Metrics(
            target_data=target_data_batch, metrics_data=[]
        )
        response = post_metrics(
            client=self._client, environment=self._environment,
            json_body=batch_request_body
        )
        return response.status_code

    def process_target(self, target_data, unique_target):
        target_attributes: List[KeyValue] = []
        if not isinstance(unique_target.attributes, Unset):
            for key, value in unique_target.attributes.items():
                # Attribute values need to be sent as string to
                # ff-server so convert all values to strings.
                target_attributes.append(KeyValue(key, str(value)))
        td = TargetData(
            identifier=unique_target.identifier,
            name=unique_target.name,
            attributes=target_attributes
        )
        target_data.append(td)

    def is_target_seen(self, target: AnalyticsEvent) -> bool:
        unique_target_key = self.get_target_key(target)

        with self._lock:
            seen = unique_target_key in self._seen_targets
        return seen

    def close(self) -> None:
        self._running = False
        if len(self._data) > 0:
            self._send_data()
        info_metrics_thread_existed()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
